# Custom VCL

You can implement your own logic for caching policy and request
processing by writing custom
[VCL](https://varnish-cache.org/docs/6.3/reference/vcl.html),
configured simply as the string value of ``spec.vcl`` in a
[VarnishConfig](/docs/ref-varnish-cfg.md#specvcl) manifest. The string
is appended "as-is" to VCL code that is generated by the controller to
implement Ingress, VarnishConfig and other features, and loaded by the
Varnish instances in your cluster. This gives you all of the
capabilites of VCL, subject to the restrictions described in the
following.

The Varnish container specialized for Ingress runs Varnish 6.3.2, and
the generated VCL code is at version 4.0. So your custom VCL must be
compatible with those versions.

For further information about VCL see:

* [vcl(7)](https://varnish-cache.org/docs/6.3/reference/vcl.html)

* [Varnish Processing States](https://varnish-cache.org/docs/6.3/reference/states.html)
  in the Varnish reference manual.

* The
  [VCL chapter](https://varnish-cache.org/docs/6.3/users-guide/vcl.html)
  in the Varnish User's Guide.

* The
  [VCL article](https://www.varnish-software.com/wiki/content/tutorials/varnish/vcl.html)
  at Varnish Software's wiki.

* The
  [VCL chapter](https://www.oreilly.com/library/view/getting-started-with/9781491972212/ch04.html)
  in Thijs Feryn's book "Getting Started with Varnish Cache".

* The
  [VCL Basics](https://book.varnish-software.com/4.0/chapters/VCL_Basics.html)
  chapter in Varnish Software's "Varnish Book" (written for Varnish
  4.0, but mostly still correct for version 6.3).

See the [``examples/`` folder](/examples/custom-vcl) for a simple
working example of custom VCL.

## Use At Your Own Risk

The correctness of the Ingress implementation, viewed broadly,
encompasses:

* Correct implementation of Ingress rules -- routing requests to
  Services based on the Host header and URL path, as specified by an
  Ingress.

* Correct implementation of VarnishConfig configurations, and other
  features documented for the project.

The VCL code generated by the controller implements these features,
and is subject to these requirements. Custom VCL could interfere with
the work of the generated code. In that case you're on your own --
broken custom VCL is not a bug of the Ingress implementation.

By following some conventions and restrictions, you can mesh custom
code smoothly with the generated code:

* The prefix ``vk8s`` (case-insensitive) is reserved by generated code
  for:

    * VCL symbol names

    * Header names

    * String values

So if you don't change anything with the ``vk8s`` prefix, you won't
change something that the generated code depends on.

* Response status values >= 60000 (for ``resp.status`` and
  ``beresp.status``) are reserved for generated code.

VCL allows you to set
[status codes >= 1000](https://varnish-cache.org/docs/6.3/reference/vcl.html#beresp) --
these are truncated to 3-digit codes in a client response, so as to form
a valid HTTP response. This is a way to control logic in VCL -- set a
special response code for which specific behaviors are implemented (for
example in ``vcl_synth`` or ``vcl_deliver``). By avoiding codes in the
reserved range, you avoid affecting the work of the generated code.

* Custom VCL is always appended *after* generated VCL.

VCL allows you to define the ``vcl_*`` subroutines
[more than once](https://varnish-cache.org/docs/6.3/reference/vcl.html#multiple-subroutines) --
they are concatenated in the order in which they appear in the VCL
source, and hence have the effect of being executed in that order,
until one of them calls ``return``, or all of them fall through. If
none of them calls ``return``, then the built-in versions of the
subroutines run (in other words,
[``builtin.vcl``](https://github.com/varnishcache/varnish-cache/blob/varnish-6.3.2/bin/varnishd/builtin.vcl)
is implicitly concatenated last).

By appending custom VCL after generated VCL, the controller ensures
that code to implement Ingress and its features is executed before
anything in custom VCL that may change the flow of control.

## Further restrictions

* It is not possible to use the
  [``include`` directive](https://varnish-cache.org/docs/6.3/reference/vcl.html#include-statement) --
  all of your custom VCL must form a single string in the VarnishConfig
  manifest.

* You cannot
  [``return(vcl(LABEL))``](https://varnish-cache.org/docs/6.3/users-guide/vcl-built-in-subs.html?highlight=label#vcl-recv)
  from ``vcl_recv``. The code is already running in a labeled VCL
  instance, and it is not possible to branch from a label to another
  label. The controller uses labels for the readiness state and to
  direct flow to the config for the current desired state, so you
  should refrain from using labels.
