# Custom VCL example

Custom VCL can be written simply by setting a string value for the
``spec.vcl`` field of a [VarnishConfig](/docs/ref-varnish-cfg.md). The
string is appended "as-is" to code generated by the controller, and
loaded by Varnish instances in the cluster. The manifest in this folder
is a simple example.

```
spec:
  vcl: |
    sub vcl_deliver {
    	set resp.http.Hello = "world";
    }

    sub vcl_backend_response {
    	set beresp.http.Backend = beresp.backend.name;
    }
```

The sample code sets the client response header ``Hello`` to "world",
and sets the backend response header ``Backend`` (which is passed
along to the client response) to the name of backend from which the
response was fetched.

Notice that the ``|`` notation is used for a block-style multiline
YAML string, indented so that all of the lines in the string are
interpreted as the value of ``vcl``. The code can of course be written
as an unformatted string (it doesn't matter to the VCL compiler), but
you may want to use one of
[YAML's multiline formats](https://yaml-multiline.info/) to write
well-formatted, readable code.

In a cluster where an Ingress and the associated Services are defined
(assume the ["cafe" example](/examples/hello) in the following), apply
the VarnishConfig definition:

```
$ kubectl apply -f custom-vcl.yaml
```

Custom VCL currently cannot be validated at apply time. Your code must
of course conform to valid VCL syntax; if not, the controller receives
error messages from the compiler when it attempts to load the code. In
that case, [Events](/docs/monitor.md) at the Warning level are
generated for the Varnish Service, including the error message from
the compiler. The error message also appears in the controller
log. Varnish usually returns status 106 for invalid VCL.

When the definition is successfully synced, it can be verified with
curl (where ``$ADDR`` is the external address of the cluster, and
``$PORT`` is the port at which requests are received by the Ingress
implementation):

```
$ curl -v -x $ADDR:$PORT http://cafe.example.com/coffee
[...]

< HTTP/1.1 200 OK
[...]
< Backend: vk8s_coffee-svc_172_17_0_8
[...]
< Hello: world
[...]

Server address: 172.17.0.8:80
Server name: coffee-6c47b9cb9c-ntv5r
Date: 07/Jan/2019:18:30:45 +0000
URI: /coffee
Request ID: b9a3abf0c2d378fe7282b614a556f46b
```

Note that the backend name in the ``Backend`` response header is
generated by the controller.

See the [docs](/docs/custom-vcl.md) for considerations about writing
custom VCL for the Ingress implementation, and for links to more
information about VCL.
