/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package haproxy

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"net/url"
	"strconv"
	"strings"
	"time"

	"github.com/go-openapi/strfmt"
	models "github.com/haproxytech/models/v2"

	"github.com/sirupsen/logrus"
)

var fmts = strfmt.NewFormats()

const (
	dataplaneUser    = "dataplaneapi"
	txPath           = "/v2/services/haproxy/transactions"
	sitesPath        = "/v2/services/haproxy/sites"
	defaultsPath     = "/v2/services/haproxy/configuration/defaults"
	reloadsPath      = "/v2/services/haproxy/reloads"
	versionHdr       = "Configuration-Version"
	reloadIDHdr      = "Reload-ID"
	varnishSock      = "unix@/varnish.sock"
	backendSock      = "unix@/run/offload/onload.sock"
	frontend         = "offloader"
	onloader         = "onloader"
	varnish          = "varnish"
	server           = varnish
	ingBackends      = "ingressBackends"
	frontendSitePath = sitesPath + "/" + frontend
	backendSitePath  = sitesPath + "/" + onloader
	crtPath          = "/etc/ssl/private"
	caBundlePath     = "/run/haproxy/ca-bundle.crt"
)

var (
	port       = int64(4443)
	sslPort    = int64(443)
	roundRobin = "roundrobin"
)

// ReloadStatus classifies the current state of a dataplane reload.
type ReloadStatus uint8

const (
	// Unknown dataplane reload status.
	Unknown ReloadStatus = iota
	// Failed dataplane reload.
	Failed
	// InProgress dataplane reload.
	InProgress
	// Succeeded dataplane reload.
	Succeeded
)

func status(reloadStr string) ReloadStatus {
	switch reloadStr {
	case "failed":
		return Failed
	case "in_progress":
		return InProgress
	case "succeeded", "success":
		return Succeeded
	default:
		return Unknown
	}
}

func (status ReloadStatus) String() string {
	switch status {
	case Failed:
		return "failed"
	case InProgress:
		return "in_progress"
	case Succeeded:
		return "succeeded"
	default:
		return "UNKNOWN"
	}
}

// ReloadState encapsulates the dataplane API's reload object.
type ReloadState struct {
	// ID is generated by the dataplane API (usually a UUID)-
	ID string
	// Response is the dataplane API's text message about the reload.
	Response string
	// Timestamp is set by the dataplane API.
	Timestamp time.Time
	// Status of the dataplane reload.
	Status ReloadStatus
}

// DataplaneError represents an error response from the dataplane
// API. Satisifies the error interface.
type DataplaneError struct {
	// Err encapsulates the dataplane API's error object.
	Err *models.Error
	// Status is the HTTP response code.
	Status int
	// Version is the configuration-version returned in the response.
	Version int
}

func (err *DataplaneError) Error() string {
	var sb strings.Builder
	sb.WriteString(http.StatusText(err.Status))
	sb.WriteString(" (")
	sb.WriteString(strconv.Itoa(err.Status))
	sb.WriteString(") version=")
	sb.WriteString(strconv.Itoa(err.Version))
	if err.Err.Code != nil {
		sb.WriteString(" models.Error.Code=")
		sb.WriteString(strconv.Itoa(int(*err.Err.Code)))
	}
	for k, v := range err.Err.Error {
		sb.WriteRune(' ')
		sb.WriteString(k)
		sb.WriteRune('=')
		sb.WriteString(v)
	}
	if err.Err.Message != nil {
		sb.WriteString(": ")
		sb.WriteString(*err.Err.Message)
	}
	return sb.String()
}

// DataplaneClient executes remote administration of a haproxy server
// using the dataplane API, which runs as a child process in the
// haproxy container.
//
//	https://www.haproxy.com/documentation/hapee/1-9r1/configuration/dataplaneapi/
//	https://www.haproxy.com/documentation/dataplaneapi/latest/
type DataplaneClient struct {
	baseURL  *url.URL
	user     string
	password string
	client   *http.Client
	log      *logrus.Logger
}

// NewDataplaneClient returns a client for the dataplane API server
// listening at host, using the Basic Auth password pass.
//
// host may have the form "addr" or "addr:port", where addr may be a
// host name or IP address.
func NewDataplaneClient(
	host, pass string, log *logrus.Logger,
) *DataplaneClient {
	return &DataplaneClient{
		baseURL: &url.URL{
			Scheme: "http",
			Host:   host,
		},
		user:     dataplaneUser,
		password: pass,
		client:   http.DefaultClient,
		log:      log,
	}
}

func (client *DataplaneClient) getHost() string {
	return client.baseURL.Host
}

var offldSite = &models.Site{
	Name: frontend,
	Service: &models.SiteService{
		Listeners: []*models.Bind{
			{
				Name:           frontend,
				Port:           &port,
				Ssl:            true,
				SslCertificate: crtPath,
			},
		},
	},
	Farms: []*models.SiteFarm{
		{
			Name:  varnish,
			UseAs: "default",
			Servers: []*models.Server{
				{
					Name:        server,
					Address:     varnishSock,
					Check:       "enabled",
					SendProxyV2: "enabled",
				},
			},
		},
	},
}

func getSite() ([]byte, error) {
	return offldSite.MarshalBinary()
}

func getOnldSite(spec *OnloadSpec) ([]byte, error) {
	// XXX: stick-table configuration
	maxConn := int64(spec.MaxConn)
	site := &models.Site{
		Name: onloader,
		Service: &models.SiteService{
			Listeners: []*models.Bind{
				{
					Name:        onloader,
					Address:     backendSock,
					Mode:        "660",
					AcceptProxy: true,
				},
			},
			Mode:    "tcp",
			Maxconn: &maxConn,
		},
		Farms: []*models.SiteFarm{
			{
				Name:  ingBackends,
				UseAs: "default",
				Mode:  "tcp",
				Balance: &models.Balance{
					Algorithm: &roundRobin,
				},
				Servers: []*models.Server{},
			},
		},
	}
	for s := 0; s < spec.Instances; s++ {
		server := &models.Server{
			Name:    fmt.Sprintf("s%02d", s),
			Address: "0.0.0.0",
			Port:    &sslPort,
			Ssl:     "enabled",
			Alpn:    "http/1.1",
			Stick:   "enabled",
		}
		if spec.Authority {
			server.Sni = "fc_pp_authority"
		}
		if spec.Verify {
			server.Verify = "required"
			server.SslCafile = caBundlePath
		} else {
			server.Verify = "none"
		}
		site.Farms[0].Servers = append(site.Farms[0].Servers, server)
	}
	return site.MarshalBinary()
}

func getDefaults(spec DefaultsSpec) ([]byte, error) {
	defaults := &models.Defaults{}
	if spec.Timeouts.Connect != nil {
		defaults.ConnectTimeout = spec.Timeouts.Connect
	}
	if spec.Timeouts.Client != nil {
		defaults.ClientTimeout = spec.Timeouts.Client
	}
	if spec.Timeouts.Server != nil {
		defaults.ServerTimeout = spec.Timeouts.Server
	}
	return defaults.MarshalBinary()
}

func drainAndClose(body io.ReadCloser) {
	io.Copy(ioutil.Discard, body)
	body.Close()
}

func (client *DataplaneClient) getReq(
	path, method string,
	body io.Reader,
	withbody bool,
) (req *http.Request, err error) {
	relPath := &url.URL{Path: path}
	url := client.baseURL.ResolveReference(relPath)
	if withbody {
		req, err = http.NewRequest(method, url.String(), body)
	} else {
		req, err = http.NewRequest(method, url.String(), nil)
	}
	if err != nil {
		return req, err
	}
	req.SetBasicAuth(client.user, client.password)
	req.Header.Set("Accept", "application/json")
	client.log.Tracef("Request: %+v", req)
	return req, nil
}

func setVersion(req *http.Request, version int64) {
	query := req.URL.Query()
	query.Add("version", strconv.FormatInt(version, 10))
	req.URL.RawQuery = query.Encode()
}

func setTx(req *http.Request, tx *models.Transaction) {
	query := req.URL.Query()
	query.Add("transaction_id", tx.ID)
	req.URL.RawQuery = query.Encode()
}

func getBody(resp *http.Response) ([]byte, error) {
	defer drainAndClose(resp.Body)
	// XXX read ContentLength bytes
	return ioutil.ReadAll(resp.Body)
}

func getTx(body []byte) (tx *models.Transaction, err error) {
	tx = &models.Transaction{}
	if err = tx.UnmarshalBinary(body); err != nil {
		err = tx.Validate(fmts)
	}
	return tx, err
}

func getError(resp *http.Response, body []byte) error {
	dplaneErr := &DataplaneError{
		Err:    &models.Error{},
		Status: resp.StatusCode,
	}
	if dplaneVer, err := strconv.Atoi(
		resp.Header.Get(versionHdr)); err != nil {
		dplaneErr.Version = dplaneVer
	}
	dplaneErr.Err.UnmarshalJSON(body)
	return dplaneErr
}

// StartTx initiates a dataplane transaction (a POST to the endpoint
// /services/haproxy/transactions) based on the given configuration
// version.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) StartTx(
	version int64) (tx *models.Transaction, err error) {

	req, err := client.getReq(txPath, http.MethodPost, nil, false)
	if err != nil {
		return
	}
	setVersion(req, version)
	resp, err := client.client.Do(req)
	if err != nil {
		return
	}
	client.log.Tracef("StartTx response: %+v", resp)
	body, err := getBody(resp)
	if err != nil {
		return
	}
	client.log.Tracef("StartTx response body: %s", string(body))

	if resp.StatusCode == http.StatusCreated {
		return getTx(body)
	}

	return nil, getError(resp, body)
}

// FinishTx completes the dataplane transaction tx. If successful, a
// haproxy configuration reload may be initiated.
//
// The DataplaneClient does not use the force-reload parameter. This
// means that dataplane does not reload the configuration
// synchronously, so as to avoid thrashing re-configuration. Reloads
// of pending transactions are performed at intervals (usually a few
// seconds).
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) FinishTx(
	tx *models.Transaction) (ReloadState, error) {

	state := ReloadState{Status: Unknown}
	req, err := client.getReq(txPath+"/"+tx.ID, http.MethodPut, nil, false)
	if err != nil {
		return state, err
	}
	setVersion(req, tx.Version)
	resp, err := client.client.Do(req)
	if err != nil {
		return state, err
	}
	client.log.Tracef("FinishTx response: %+v", resp)
	body, err := getBody(resp)
	if err != nil {
		return state, err
	}
	client.log.Tracef("FinishTx response body: %s", string(body))

	switch resp.StatusCode {
	case http.StatusOK, http.StatusAccepted:
		if tx, err = getTx(body); err != nil {
			return state, err
		}
		if resp.StatusCode == http.StatusOK {
			state.Status = Succeeded
			state.Timestamp = time.Now()
			return state, nil
		}
		state.ID = resp.Header.Get(reloadIDHdr)
		state.Status = status(tx.Status)
		if state.Status == Succeeded {
			state.Timestamp = time.Now()
		}
		return state, nil
	default:
		return state, getError(resp, body)
	}
}

func (client *DataplaneClient) configTLS(
	tx *models.Transaction,
	path, method string,
	onldSpec *OnloadSpec,
) error {
	var rdr *bytes.Reader
	if method != http.MethodDelete {
		var siteBytes []byte
		var err error

		if onldSpec == nil {
			siteBytes, err = getSite()
		} else {
			siteBytes, err = getOnldSite(onldSpec)
		}
		if err != nil {
			return err
		}
		rdr = bytes.NewReader(siteBytes)
	}

	req, err := client.getReq(path, method, rdr,
		method != http.MethodDelete)
	if err != nil {
		return err
	}
	setTx(req, tx)
	if method != http.MethodDelete {
		req.Header.Set("Content-Type", "application/json")
	}

	resp, err := client.client.Do(req)
	if err != nil {
		return err
	}
	client.log.Tracef("configTLS response: %+v", resp)

	body, err := getBody(resp)
	if err != nil {
		return err
	}
	client.log.Tracef("configTLS response body: %s", string(body))

	switch resp.StatusCode {
	case http.StatusAccepted:
		// XXX return the Reload-ID?
		if method != http.MethodDelete {
			// Validate the response body
			site := &models.Site{}
			if err = site.UnmarshalBinary(body); err != nil {
				return err
			}
			if err = site.Validate(fmts); err != nil {
				return err
			}
		}
		return nil
	case http.StatusOK, http.StatusCreated, http.StatusNoContent:
		panic("Got " + http.StatusText(resp.StatusCode) +
			" but did not set force_reload")
	default:
		return getError(resp, body)
	}
}

func (client *DataplaneClient) configDefaults(
	tx *models.Transaction,
	spec DefaultsSpec,
) error {
	defaultBytes, err := getDefaults(spec)
	if err != nil {
		return err
	}
	rdr := bytes.NewReader(defaultBytes)

	req, err := client.getReq(defaultsPath, http.MethodPut, rdr, true)
	if err != nil {
		return err
	}
	setTx(req, tx)
	req.Header.Set("Content-Type", "application/json")

	resp, err := client.client.Do(req)
	if err != nil {
		return err
	}
	client.log.Tracef("configDefaults response: %+v", resp)

	body, err := getBody(resp)
	if err != nil {
		return err
	}
	client.log.Tracef("configDefaults response body: %s", string(body))

	switch resp.StatusCode {
	case http.StatusAccepted:
		// XXX return the Reload-ID?
		// Validate the response body
		defaults := &models.Defaults{}
		if err = defaults.UnmarshalBinary(body); err != nil {
			return err
		}
		if err = defaults.Validate(fmts); err != nil {
			return err
		}
		return nil
	case http.StatusOK, http.StatusCreated, http.StatusNoContent:
		panic("Got " + http.StatusText(resp.StatusCode) +
			" but did not set force_reload")
	default:
		return getError(resp, body)
	}
}

// AddOffldr adds the offloader configuration for haproxy, in the
// dataplane transaction tx.
//
// AddOffldr MUST be used if the offloader was not configured previously
// since the haproxy container was started, or after deletion.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) AddOffldr(tx *models.Transaction) error {
	return client.configTLS(tx, sitesPath, http.MethodPost, nil)
}

// UpdateOffldr modifies the offloader configuration for haproxy, in
// the dataplane transaction tx.
//
// UpdateOffldr MUST be used if the offloader was previously added with
// AddOffldr, and not removed with DeleteOffldr.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) UpdateOffldr(tx *models.Transaction) error {
	return client.configTLS(tx, frontendSitePath, http.MethodPut, nil)
}

// DeleteOffldr removes the haproxy offloader configuration, in the
// dataplane transaction tx.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) DeleteOffldr(tx *models.Transaction) error {
	return client.configTLS(tx, frontendSitePath, http.MethodDelete, nil)
}

// AddOnldr adds the onloader configuration for haproxy, in the
// dataplane transaction tx. instances specifies the number of servers
// in the haproxy backend.
//
// AddOnldr MUST be used if the onloader was not configured previously
// since the haproxy container was started, or after deletion.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) AddOnldr(
	tx *models.Transaction,
	onldSpec *OnloadSpec,
) error {
	return client.configTLS(tx, sitesPath, http.MethodPost, onldSpec)
}

// UpdateOnldr modifies the onloader configuration for haproxy, in the
// dataplane transaction tx. instances specifies the number of servers
// in the haproxy backend.
//
// UpdateOnldr MUST be used if the onloader was previously added with
// AddOnldr, and not removed with DeleteOnldr.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) UpdateOnldr(
	tx *models.Transaction,
	onldSpec *OnloadSpec,
) error {
	return client.configTLS(tx, backendSitePath, http.MethodPut, onldSpec)
}

// DeleteOnldr removes the haproxy onloader configuration, in the
// dataplane transaction tx.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) DeleteOnldr(tx *models.Transaction) error {
	return client.configTLS(tx, backendSitePath, http.MethodDelete, nil)
}

// UpdateDefaults modifies default haproxy configuration (valid for both
// off- and onload).
func (client *DataplaneClient) UpdateDefaults(
	tx *models.Transaction,
	spec *DefaultsSpec,
) error {
	if spec == nil {
		return nil
	}
	return client.configDefaults(tx, *spec)
}

// DeleteTx removes the dataplane transaction tx. This should be
// called after a successful invocation of FinishTx(tx).
//
// A non-nil error return may wrap a DataplaneError.
//
// XXX currently a no-op.
func (client *DataplaneClient) DeleteTx(tx *models.Transaction) error {
	return nil
}

// Reloaded returns true if the reload identified by id has been
// successfully complete, with details about the reload state.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) Reloaded(id string) (bool, ReloadState, error) {
	state := ReloadState{ID: id}
	req, err := client.getReq(reloadsPath+"/"+id, http.MethodGet, nil,
		false)
	if err != nil {
		return false, state, err
	}
	resp, err := client.client.Do(req)
	if err != nil {
		return false, state, err
	}
	client.log.Tracef("Reloaded response: %+v", resp)
	body, err := getBody(resp)
	if err != nil {
		return false, state, err
	}
	client.log.Tracef("Reloaded response body: %s", string(body))

	switch resp.StatusCode {
	case http.StatusOK:
		reload := &models.Reload{}
		if err = reload.UnmarshalBinary(body); err != nil {
			return false, state, err
		}
		if err = reload.Validate(fmts); err != nil {
			return false, state, err
		}
		state.ID = reload.ID
		state.Response = reload.Response
		state.Timestamp = time.Unix(reload.ReloadTimestamp, 0)
		state.Status = status(reload.Status)
		return true, state, nil
	default:
		dplaneErr := getError(resp, body)
		if resp.StatusCode == http.StatusNotFound {
			return false, state, nil
		}
		return false, state, dplaneErr
	}
}

type sitesBody struct {
	Version int          `json:"_version,omitempty"`
	Sites   models.Sites `json:"data"`
}

// LoaderStatus returns booleans for whether the off- and onloader
// sites have been loaded by the dataplane API, and returns the
// current configuration version.
//
// A non-nil error return may wrap a DataplaneError.
func (client *DataplaneClient) LoaderStatus() (
	offLoaded, onLoaded bool, version int, err error) {

	req, err := client.getReq(sitesPath, http.MethodGet, nil, false)
	if err != nil {
		return
	}
	resp, err := client.client.Do(req)
	if err != nil {
		return
	}
	client.log.Tracef("dataplane sites response: %+v", resp)
	body, err := getBody(resp)
	if err != nil {
		return
	}
	client.log.Tracef("dataplane sites response body: %s", string(body))

	if verStr := resp.Header.Get(versionHdr); verStr != "" {
		client.log.Tracef("dataplane sites version: %s", verStr)
		if version, err = strconv.Atoi(verStr); err != nil {
			return
		}
	}

	if resp.StatusCode != http.StatusOK {
		err = getError(resp, body)
		return
	}

	var sb sitesBody
	if err = json.Unmarshal(body, &sb); err != nil {
		return
	}
	if err = sb.Sites.Validate(fmts); err != nil {
		return
	}
	if version == 0 {
		version = sb.Version
	}
	// XXX what if version from body & the header don't match?
	for _, site := range []*models.Site(sb.Sites) {
		if site.Name == frontend {
			offLoaded = true
		} else if site.Name == onloader {
			onLoaded = true
		}
		if offLoaded && onLoaded {
			break
		}
	}
	return
}
